@@ -69,23 +69,11 @@ class LocalLinear:
         The fitted values for the response variable.
     residuals : np.ndarray
         The residuals resulting from the local linear regression.
-    
-    
-
-    Methods
-    -------
-    fit()
-        Fit the model by local linear estimation and return estimated coefficients.
-    summary()
-        Print a summary of local linear regression results, including bandwidth, number of observations, and beta coefficients.
-    plot_betas()
-        Plot estimated coefficients curve over a normalized x-axis from 0 to 1.
-    plot_predicted()
-        Plot true data against fitted values.
-    plot_residuals()
-        Plot residuals.
-    confidence_bands()
-        Construct and plot bootstrap confidence intervals/bands for each coefficient curve.
+        
+    Raises
+    ------
+    ValueError
+        No valid bandwidth selection procedure is provided.
 
     Notes
     -----
@@ -107,7 +95,6 @@ class LocalLinear:
         mX: np.ndarray,
         h: float = 0,
         bw_selection: str = None,
-        tau: np.ndarray = None,
         kernel: str = "epanechnikov",
         LB_bw: float = None,
         UB_bw: float = None
@@ -118,19 +105,9 @@ class LocalLinear:
         self.n = len(vY)
         self.times = np.arange(1 / self.n, ( self.n + 1) /  self.n, 1 /  self.n)
 
-        if tau is None:
-            self.tau = np.arange(1/self.n,(self.n+1)/self.n,1/self.n)
-            self.tau_index=np.array([0,self.n])
-        elif isinstance(tau, np.ndarray):
-            if tau[0] <= 0:
-                self.tau = np.arange(1 / self.n, (self.n + 1) / self.n, 1 / self.n)[
-                           int(0):int(tau[1] * self.n)]
+        self.tau = np.arange(1 / self.n, (self.n + 1) / self.n, 1 / self.n)
+        # self.tau_index = np.array([0, self.n])
 
-                self.tau_index = np.array([int(0), int(tau[1] * self.n)])
-            else:
-                self.tau = np.arange(1/self.n,(self.n+1)/self.n,1/self.n)[int(tau[0]*self.n-1):int(tau[1]*self.n)]
-
-                self.tau_index = np.array([int(tau[0]*self.n-1),int(tau[1]*self.n)])
         # elif isinstance(tau, float):
         #     self.tau = np.array([tau])
         self.tau_bw_selection = np.arange(1/self.n,(self.n+1)/self.n,1/self.n)
@@ -158,9 +135,8 @@ class LocalLinear:
 
             if self.bw_selection not in ['all', 'aic', 'gcv']:
                 if self.bw_selection[:4] != 'lmcv':
-                    print(
-                        'Error: bandwidth selection method is invalid. \nPlease provide an expression from the following options: [\'all\', \'aic\', \'gcv\'] or use \'lmcv_l\'')
-                    exit(1)
+                    raise ValueError(
+                        'Bandwidth selection method is invalid. \nPlease provide an expression from the following options: [\'all\', \'aic\', \'gcv\'] or use \'lmcv_l\'')
                 else:
                     self.lmcv_type=self.bw_selection[5]
 
@@ -169,8 +145,9 @@ class LocalLinear:
 
 
             if bw_selection is None:
-                self.dict_bw = self.bandwidth_selection(LB_bw, UB_bw)
-                print('-----------------------------------------------------------')
+                self.dict_bw = self._bandwidth_selection(LB_bw, UB_bw)
+                print('\n')
+                print("=" * 60)
                 print('Optimal bandwidth selected by individual method:')
                 print('- AIC method:', format(self.dict_bw['aic'], '.4f'))
                 print('- GCV method: ', format(self.dict_bw['gcv'], '.4f'))
@@ -178,32 +155,33 @@ class LocalLinear:
                 print('- LMCV-2 method: ', format(self.dict_bw['2'],'.4f'))
                 print('- LMCV-4 method: ', format(self.dict_bw['4'],'.4f'))
                 print('- LMCV-6 method: ', format(self.dict_bw['6'],'.4f'))
-                print('-----------------------------------------------------------')
+                print("=" * 60)
                 self.dict_bw['all'] = np.array(list(self.dict_bw.values())).mean()
                 self.h = self.dict_bw['all']
                 self.h_gcv = self.dict_bw['gcv']
-                print(f'Optimal bandwidth used is the avg. of all methods: {self.h: .4f}\n')
-                print('-----------------------------------------------------------')
+                print(f'Optimal bandwidth used is the avg. of all methods: {self.h: .4f}')
+                print("=" * 60)
                 print(
-                    f'Note: For constructing confidence intervals/bands using the residual-based bootstrap method, the avg. bandwidth of all methods {self.h: .4f} is used.\n')
-                print(f'For constructing confidence intervals/bands using the MB method, the GCV bandwidth {self.h_gcv: .4f} is used.\n')
+                    f'Note: (1) For constructing confidence intervals/bands using the residual-based bootstrap method, the avg. bandwidth of all methods {self.h: .4f} is used; (2) For constructing confidence intervals/bands using the MB method, the GCV bandwidth {self.h_gcv: .4f} is used.')
                 # print(f'If the MB is implemented, the GCV bandwidth {self.the_gcv_h: .4f} is used.')
 
 
             elif self.bw_selection[:4] == 'lmcv':
-                self.h = self.bandwidth_selection(LB_bw, UB_bw)
+                self.h = self._bandwidth_selection(LB_bw, UB_bw)
+                print('\n')
                 print(f'- LMCV-{self.lmcv_type} method: ', format(self.h,'.4f'))
-                print(f'Optimal bandwidth used is {self.bw_selection}: {self.h: .4f}\n')
-                print('-----------------------------------------------------------')
-                print(f'Note: For constructing confidence intervals/bands using the MB method, a GCV bandwidth is recommeneded.\n')
+                print(f'Optimal bandwidth used is {self.bw_selection}: {self.h: .4f}')
+                print("=" * 60)
+                print('Note: For constructing confidence intervals/bands using the MB method, a GCV bandwidth is recommended.\n')
                 # print(
                 #     f'Note: If a residual-based bootstrap method (LBWB, WB, SB, SWB, AWB) is adopted, the {self.bw_selection} bandwidth {self.h: .4f} is used.')
                 # print(f'If the MB is implemented, the GCV bandwidth {self.the_gcv_h: .4f} is used.')
 
             # self.h = self.dict_bw[self.bw_selection]
             elif self.bw_selection == 'all':
-                self.dict_bw = self.bandwidth_selection(LB_bw, UB_bw)
-                print('-----------------------------------------------------------')
+                self.dict_bw = self._bandwidth_selection(LB_bw, UB_bw)
+                print('\n')
+                print("=" * 60)
                 print('Optimal bandwidth selected by individual method:')
                 print('- AIC method:', format(self.dict_bw['aic'], '.4f'))
                 print('- GCV method: ', format(self.dict_bw['gcv'], '.4f'))
@@ -211,25 +189,25 @@ class LocalLinear:
                 print('- LMCV-2 method: ', format(self.dict_bw['2'], '.4f'))
                 print('- LMCV-4 method: ', format(self.dict_bw['4'], '.4f'))
                 print('- LMCV-6 method: ', format(self.dict_bw['6'], '.4f'))
-                print('-----------------------------------------------------------')
+                print("=" * 60)
                 self.dict_bw['all'] = np.array(list(self.dict_bw.values())).mean()
                 self.h = self.dict_bw['all']
-                print(f'Optimal bandwidth used is the avg. of all methods: {self.h: .4f}\n')
-                print('-----------------------------------------------------------')
-                print('Note: For constructing confidence intervals/bands using the MB method, a GCV bandwidth is recommeneded.\n')
+                print(f'Optimal bandwidth used is the avg. of all methods: {self.h: .4f}')
+                print("=" * 60)
+                print('Note: For constructing confidence intervals/bands using the MB method, a GCV bandwidth is recommended.\n')
                 # print(f'If the MB is implemented, the GCV bandwidth {self.the_gcv_h: .4f} is used.')
 
             elif self.bw_selection == 'gcv' or 'aic':
-                self.h  = self.bandwidth_selection(LB_bw, UB_bw)
-                print(f'Optimal bandwidth used is {self.bw_selection}: {self.h: .4f}\n')
-                print('-----------------------------------------------------------')
+                self.h  = self._bandwidth_selection(LB_bw, UB_bw)
+                print(f'Optimal bandwidth used is {self.bw_selection}: {self.h: .4f}')
+                print("=" * 60)
                 if self.bw_selection == 'gcv':
                     print(
-                        'Note: For constructing confidence intervals/bands using the residual-based bootstrap method, a LMCV bandwidth is recommeneded.\n')
+                        'Note: For constructing confidence intervals/bands using the residual-based bootstrap method, a LMCV bandwidth is recommended.\n')
 
                 else:
                     print(
-                        'Note: For constructing confidence intervals/bands using the MB method, a GCV bandwidth is recommeneded.\n')
+                        'Note: For constructing confidence intervals/bands using the MB method, a GCV bandwidth is recommended.\n')
 
         else:
             self.h = h
@@ -248,6 +226,11 @@ class LocalLinear:
         ----------
         u : np.ndarray
             An array of scaled distances.
+            
+        Raises
+        ------
+        ValueError
+            No valid kernel name is provided.
 
         Returns
         -------
@@ -509,12 +492,12 @@ class LocalLinear:
             predicted_y = betahat[0]
         else:
             # Multiple betas
-            predicted_y = np.sum(betahat * self.mX[self.tau_index[0]: self.tau_index[1]].T, axis=0)
+            predicted_y = np.sum(betahat * self.mX.T, axis=0)
             
         # store results in corresponding attributes
         self.betahat = betahat
         self.predicted_y = (predicted_y.reshape(-1,1))
-        self.residuals = (self.vY)[self.tau_index[0]:self.tau_index[1]] - self.predicted_y
+        self.residuals = (self.vY) - self.predicted_y
 
         return self.betahat
 
@@ -642,7 +625,8 @@ class LocalLinear:
         vh = np.arange(LB_bw, UB_bw, 0.005)
         
         betasss = np.zeros(shape=(len(vh), self.n_est, self.n))
-        for index, h in enumerate(vh):
+        print('Progress LMCV:')
+        for index, h in enumerate(tqdm(vh)):
             betasss[index] = self._beta_estimation_lmcv(
                 h, tau=self.times, lmcv_type=lmcv_type)
 
@@ -675,7 +659,9 @@ class LocalLinear:
         if self.bw_selection != 'all':
             return self._get_optimalh_lmcv(self.lmcv_type, LB_bw, UB_bw)
         h = []
-        for lmcv_type in [0, 2, 4, 6]:
+        lmcv_types = [0, 2, 4, 6]
+        for r in range(len(lmcv_types)):
+            lmcv_type = lmcv_types[r]
             h.append(self._get_optimalh_lmcv(lmcv_type, LB_bw, UB_bw))
         AVG = np.mean(h)
         h.append(AVG)
@@ -683,7 +669,7 @@ class LocalLinear:
             h.append(self._get_optimalh_lmcv(int(self.lmcv_type), LB_bw, UB_bw))
         return h
 
-    def AICmodx(self, h):
+    def _AICmodx(self, h):
         '''
         Computes the AIC value for a given mean squared error and trace.
 
@@ -828,12 +814,12 @@ class LocalLinear:
             Optimal bandwidth value
 
         '''
-        res = scipy.optimize.minimize(self.AICmodx, 0.1, bounds=[(LB_bw, UB_bw)])
+        res = scipy.optimize.minimize(self._AICmodx, 0.1, bounds=[(LB_bw, UB_bw)])
         h_opt = res.x[0]
 
         return h_opt
 
-    def GCVmodx(self, h):
+    def _GCVmodx(self, h):
         '''
         Computes the GCV value for a bandwidth value.
 
@@ -868,12 +854,12 @@ class LocalLinear:
             Optimal bandwidth value
 
         '''        
-        res = scipy.optimize.minimize(self.GCVmodx, 0.1, bounds=[(LB_bw, UB_bw)])
+        res = scipy.optimize.minimize(self._GCVmodx, 0.1, bounds=[(LB_bw, UB_bw)])
         h_opt = res.x[0]
 
         return h_opt
 
-    def bandwidth_selection(self, LB_bw, UB_bw):
+    def _bandwidth_selection(self, LB_bw, UB_bw):
         """
         Calculate the optimal bandwidth for the local linear regression model using LMCV, AIC and GCV.
         
@@ -917,7 +903,7 @@ class LocalLinear:
     # Bootstrap
     ############################################################################################################
 
-    def AR(self, zhat, T, ic=None):
+    def _AR(self, zhat, T, ic=None):
         """
         Estimate the AR model and compute residuals.
 
@@ -990,7 +976,7 @@ class LocalLinear:
             zstar = np.append(zstar, ar_component)
         return zstar
 
-    def S_BT(
+    def _S_BT(
         self,
         epsilontilde: np.ndarray,
         max_lag: int,
@@ -1001,7 +987,7 @@ class LocalLinear:
     ):
         """
         Sieve Bootstrap
-        Calculate the transformed response variable using the local linear regression model.
+        Compute a bootstrap sample using the sieve bootstrap.
 
         Parameters
         ----------
@@ -1021,7 +1007,7 @@ class LocalLinear:
         Returns
         -------
         vYstar : np.ndarray
-            Transformed response variable.
+            The bootstrap sample generated by the _S_BT algorithm.
         """
         epsilonstar = np.random.choice(epsilontilde, T - max_lag + 50)
         if max_lag == 0:
@@ -1039,7 +1025,7 @@ class LocalLinear:
             return vYstar
 
     ######### Autoregressive Bootstrap #########
-    def AW_BT(
+    def _AW_BT(
         self,
         zhat: np.ndarray,
         mX: np.ndarray,
@@ -1070,7 +1056,7 @@ class LocalLinear:
         Returns
         -------
         vYstar : np.ndarray
-            Transformed response variable.
+            The bootstrap sample generated by the _AW_BT algorithm.
         """
         # v_nv_star=np.random.normal(0,np.sqrt(1-gamma**2),200)
         xi_star0 = np.random.normal(0, 1, 1)
@@ -1087,7 +1073,7 @@ class LocalLinear:
 
         return vYstar
 
-    def W_BT(
+    def _W_BT(
         self,
         zhat: np.ndarray,
         mX: np.ndarray,
@@ -1098,7 +1084,7 @@ class LocalLinear:
     ):
         """
         Wild Bootstrap
-        Calculate the transformed response variable using the local linear regression model.
+        Compute a bootstrap sample using the wild bootstrap.
 
         Parameters
         ----------
@@ -1118,7 +1104,7 @@ class LocalLinear:
         Returns
         -------
         vYstar : np.ndarray
-            Transformed response variable.
+            The bootstrap sample generated by the _W_BT algorithm.
         """
         # generate zstar by
         zstar = zhat * np.random.normal(0, 1, T)
@@ -1130,7 +1116,7 @@ class LocalLinear:
             vYstar = (mX @ betatilde + zstar).diagonal()
             return vYstar
 
-    def SW_BT(
+    def _SW_BT(
         self,
         epsilontilde: np.ndarray,
         max_lag: int,
@@ -1141,7 +1127,7 @@ class LocalLinear:
     ):
         """
         Sieve Wild Bootstrap
-        Calculate the transformed response variable using the local linear regression model.
+        Compute a bootstrap sample using the sieve wild bootstrap.
 
         Parameters
         ----------
@@ -1161,7 +1147,7 @@ class LocalLinear:
         Returns
         -------
         vYstar : np.ndarray
-            Transformed response variable.
+            The bootstrap sample generated by the _SW_BT algorithm.
         """
         epsilonstar = epsilontilde * np.random.normal(0, 1, T - max_lag)
         epsilonstar = np.random.choice(epsilonstar, T - max_lag + 50)
@@ -1179,12 +1165,12 @@ class LocalLinear:
             vYstar = (mX @ betatilde + zstar_array).diagonal()
             return vYstar
 
-    def LBW_BT(
+    def _LBW_BT(
         self, zhat: np.ndarray, mX: np.ndarray, betatilde: np.ndarray, T: int, h: float, gamma: float
     ):
         """
         Local Blockwise Wild Bootstrap
-        Performs the local blockwise wild bootstrap algorithm to generate a bootstrap sample.
+        Compute a bootstrap sample using the local blockwise wild bootstrap.
 
         Parameters
         ----------
@@ -1204,7 +1190,7 @@ class LocalLinear:
         Returns
         -------
         vYstar : np.ndarray
-            The bootstrap sample generated by the LBW_BT algorithm.
+            The bootstrap sample generated by the _LBW_BT algorithm.
         """
         l = int(4.5 * ((T * h) ** (0.25)))
         number_blocks = T - l + 1
@@ -1269,7 +1255,7 @@ class LocalLinear:
             qtau[1, i] = np.quantile(diff[:, i], (1 - alphap / 2))
         return qtau
 
-    def ABS_value(self, qtau, diff, tau, alpha, B):
+    def _ABS_value(self, qtau, diff, tau, alpha, B):
         """
         Calculate the absolute value for quantiles.
 
@@ -1300,7 +1286,7 @@ class LocalLinear:
 
         return np.abs((np.sum(np.where(check == len(tau), 1, 0)) / B) - (1 - alpha))
 
-    def min_alphap(self, diff, tau, alpha, B):
+    def _min_alphap(self, diff, tau, alpha, B):
         """
         Calculate the minimum alpha value for confidence bands.
 
@@ -1321,11 +1307,11 @@ class LocalLinear:
             Minimum alpha value.
         """
         
-        last = self.ABS_value(self._get_qtau(
+        last = self._ABS_value(self._get_qtau(
             1 / B, diff, tau), diff, tau, alpha, B)
         for index, alphap in enumerate(np.arange(2, int(B * alpha) + 2) / B):
             qtau = self._get_qtau(alphap, diff, tau)
-            value = self.ABS_value(qtau, diff, tau, alpha, B)
+            value = self._ABS_value(qtau, diff, tau, alpha, B)
             if value <= last:
                 last = value
                 if index == int(B * alpha) - 1:
@@ -1336,7 +1322,7 @@ class LocalLinear:
                 else:
                     return (index) / B
 
-    def K_ZW_Q(self, vTi_t, h):
+    def _K_ZW_Q(self, vTi_t, h):
         """
         Kernel function for Multiplier Bootstrap quantile.
 
@@ -1354,7 +1340,7 @@ class LocalLinear:
         """
         return 2 * np.sqrt(2) * self._kernel(np.sqrt(2) * vTi_t / h) - self._kernel(vTi_t / h)
 
-    def getQ(self, B, T, taut, h, alpha):
+    def _getQ(self, B, T, taut, h, alpha):
         """
         Calculate quantile for Multiplier Bootstrap samples.
 
@@ -1389,13 +1375,13 @@ class LocalLinear:
                 vMu = mMu[:, i]
                 vTi_t = mTi_t[:, t - 1]
                 vV = mV[:, i]
-                vKernel = self.K_ZW_Q(vTi_t, h)
+                vKernel = self._K_ZW_Q(vTi_t, h)
                 vMu[t - 1] = vV @ vKernel
             vQ[i] = max(np.absolute(vMu)) / dDenum
         dQ = np.quantile(vQ, 1 - alpha)
         return dQ
 
-    def getLambda(self, mmDelta, iN, iM, taut, dT, dTau, dGamma):
+    def _getLambda(self, mmDelta, iN, iM, taut, dT, dTau, dGamma):
         """
         Calculate lambda values for Multiplier Bootstrap samples.
 
@@ -1434,7 +1420,7 @@ class LocalLinear:
             mLam = mLam + vWeights[i] * mmDelta[:, :, i]
         return mLam
 
-    def getDelta(self, mX, vEps, iM):
+    def _getDelta(self, mX, vEps, iM):
         """
         Calculate delta values for Multiplier Bootstrap samples.
 
@@ -1467,7 +1453,7 @@ class LocalLinear:
             mmDelta[:, :, i] = mDeltai
         return mmDelta
 
-    def getSigma(self, mM, mLam):
+    def _getSigma(self, mM, mLam):
         """
         Calculate sigma values for Multiplier Bootstrap samples.
 
@@ -1488,7 +1474,7 @@ class LocalLinear:
                           for d in range(mLam.shape[0])])
         return vSigma
 
-    def getM(self, mX, T, dT, h):
+    def _getM(self, mX, T, dT, h):
         """
         Calculate M values for Multiplier Bootstrap samples.
 
@@ -1516,7 +1502,7 @@ class LocalLinear:
         mM = mX_tilde.T @ mX_tilde / (T * h)
         return mM
 
-    def MC_ZW(self, alpha, h, vY, mX, T, B):
+    def _MC_ZW(self, alpha, h, vY, mX, T, B):
         """
         Perform the Multiplier Bootstrap bootstrap algorithm.
         Zhou, Z., & Wu, W. B. (2010). Simultaneous inference of linear models with time varying coefficients.
@@ -1558,14 +1544,14 @@ class LocalLinear:
         # Jackknife bias-corrected estimator
         h_jk = 2 * h
         betahat_j = self._est_betas(vY, mX, h, taut, taut, self.n_est)
-        betahat_k = self._est_betas(vY, mX, h_jk, taut, taut, self.n_est)
+        betahat_k = self._est_betas(vY, mX, h_jk/np.sqrt(2), taut, taut, self.n_est)
         betahat_jk = 2 * betahat_k - betahat_j
 
         # Simulation quantile
-        dQ = self.getQ(B, T, taut, h, alpha)
+        dQ = self._getQ(B, T, taut, h, alpha)
 
         vEps = vY - (mX @ betahat).diagonal()
-        mmDelta = self.getDelta(mX, vEps, iM)
+        mmDelta = self._getDelta(mX, vEps, iM)
 
         vT_star = np.maximum(np.array([h] * T), np.minimum(taut, 1 - h))
 
@@ -1573,13 +1559,13 @@ class LocalLinear:
 
         for l in range(T):
             # Construct mMHat(tau)
-            mM = self.getM(mX, T, vT_star[l], h)
+            mM = self._getM(mX, T, vT_star[l], h)
 
             # Construct mLambdaTilde(tau)
-            mLam = self.getLambda(mmDelta, T, iM, taut, taut[l], dTau, dGamma)
+            mLam = self._getLambda(mmDelta, T, iM, taut, taut[l], dTau, dGamma)
 
             # Construct LRV
-            vSigma = self.getSigma(mM, mLam)
+            vSigma = self._getSigma(mM, mLam)
             for d in range(mX.shape[1]):
                 mmSCT[l, :, d] = betahat_jk[d, l] + \
                     vSigma[d] * dQ * np.array([-1, 1])
@@ -1597,7 +1583,7 @@ class LocalLinear:
         results.append(betahat_jk)
         return results
 
-    def construct_confidence_bands(self, bootstraptype: str, alpha: float = None, gamma: float = None, ic: str = None, Gsubs: list = None, Chtilde: float = None, B: float = 1299):
+    def _construct_confidence_bands(self, bootstraptype: str, alpha: float = None, gamma: float = None, ic: str = None, Gsubs: list = None, Chtilde: float = None, B: float = 1299):
         """
         Construct confidence bands using bootstrap methods.
 
@@ -1621,6 +1607,11 @@ class LocalLinear:
         B : int
             The number of bootstrap samples.
             Deafult is 1299, if not provided by the user.
+            
+        Raises
+        ------
+        ValueError
+            No valid bootstrap type is provided.
 
         Returns
         -------
@@ -1641,7 +1632,7 @@ class LocalLinear:
 
         if bootstraptype == "MB":
             print("Calculating MB Bootstrap Samples")
-            return self.MC_ZW(alpha, self.h, self.vY, self.mX, len(self.vY), B)
+            return self._MC_ZW(alpha, self.h, self.vY, self.mX, len(self.vY), B)
 
         if Chtilde is None or Chtilde <= 0:
             Chtilde = 2
@@ -1658,11 +1649,11 @@ class LocalLinear:
 
         # Determine the appropriate bootstrap function
         bootstrap_functions = {
-            "SB": self.S_BT,
-            "WB": self.W_BT,
-            "SWB": self.SW_BT,
-            "LBWB": self.LBW_BT,
-            "AWB": self.AW_BT
+            "SB": self._S_BT,
+            "WB": self._W_BT,
+            "SWB": self._SW_BT,
+            "LBWB": self._LBW_BT,
+            "AWB": self._AW_BT
         }
         
         if B < 1299:
@@ -1693,7 +1684,7 @@ class LocalLinear:
         print(f"Calculating {bootstraptype} Bootstrap Samples")
         for i in tqdm(range(B)):
             if bootstraptype in ["SB", "SWB"]:
-                epsilonhat, max_lag, armodel = self.AR(zhat, T, ic)
+                epsilonhat, max_lag, armodel = self._AR(zhat, T, ic)
                 epsilontilde = epsilonhat - np.mean(epsilonhat)
                 if bootstraptype == 'SWB':
                     vYstar = bootstrap_function(
@@ -1725,7 +1716,7 @@ class LocalLinear:
                 for k in range(self.n_est):
                     diff_beta_G[k][i] = diff_G[k]
 
-            optimal_alphap_G = [self.min_alphap(
+            optimal_alphap_G = [self._min_alphap(
                 diff_beta_G[i], G, alpha, B) for i in range(self.n_est)]
 
             S_LB_beta = [betahat_G[i] - self._get_qtau(optimal_alphap_G[i], diff_beta_G[i], G)[
@@ -1747,35 +1738,66 @@ class LocalLinear:
         Print a summary of the regression results.
         """
         print("Local Linear Regression Results")
-        print("=" * 30)
-        print(f"Bandwidth: {self.h: .4f}")
+        print("=" * 60)
+        print(f'Kernel: {self.kernel}')
+        if self.bw_selection is None:
+            print(f'Bandwidth specified by user: {self.h: .4f}')
+        elif self.bw_selection == 'all':
+            print('Bandwidth selection method: AVG')
+            print(f"Bandwidth used for estimation: {self.h: .4f}")
+        else:
+            print(f'Bandwidth selection method: {self.bw_selection.upper()}')
+            print(f"Bandwidth used for estimation: {self.h: .4f}")
         print(f"Number of observations: {len(self.vY)}")
         print(f"Number of predictors: {self.betahat.shape[0]}")
-        print("=" * 30)
+        print("=" * 60)
         print(f"Beta coefficients (shape: {self.betahat.shape}):")
         print("Use the 'plot_betas()' method to plot the beta coefficients.")
-        print("=" * 30)
+        print("=" * 60)
         print(
             "Use the 'confidence_bands()' method to obtain the confidence bands and plots.")
         print(
             "You can choose out of 6 types of Bootstrap to construct confidence bands:")
         print("SB, WB, SWB, MB, LBWB, AWB")
-        print("=" * 30)
-        print("Use the 'plot_residuals()' method to plot the residuals.")
+        print("=" * 60)
+        print("Use the 'plot_residuals()' method to plot the residuals.\n")
 
 
-    def plot_betas(self):
+    def plot_betas(self, tau: list = None):
 
         """
         Plot the beta coefficients over a normalized x-axis from 0 to 1.
+        
+        Parameters
+        ----------
+        tau : list, optional
+            The list looks the  following: tau = [start,end].
+            The function will plot all data and estimates between start and end.
+            
+        Raises
+        ------
+        ValueError
+            No valid tau is provided.
+        
         """
-        x_vals = np.arange(1/self.n,(self.n+1)/self.n,1/self.n)[self.tau_index[0]: self.tau_index[1]]
+        tau_index = None
+        if tau is None:
+            tau_index=np.array([0,self.n])
+        elif isinstance(tau, list):
+            if min(tau) <= 0:
+                tau_index = np.array([int(0), int(max(tau) * self.n)])
+            else:
+                tau_index = np.array([int(min(tau)*self.n-1),int(max(tau)*self.n)])
+        else:
+            raise ValueError('The optional parameter tau is required to be a list.')
+            
+        x_vals = np.arange(1/self.n,(self.n+1)/self.n,1/self.n)
         
         if self.n_est == 1:
     
             plt.figure(figsize=(12, 6))
-            plt.plot(x_vals, self.vY[self.tau_index[0]:self.tau_index[1]], label="True data", linewidth=2,color='black')
-            plt.plot(x_vals, self.betahat.reshape(-1,1), label="Estimated $\\beta_{0}$", linestyle="--", linewidth=2)
+            plt.plot(x_vals[tau_index[0]:tau_index[1]], self.vY[tau_index[0]:tau_index[1]], label="True data", linewidth=1,color='black')
+            plt.plot(x_vals[tau_index[0]:tau_index[1]], self.betahat.reshape(-1,1)[tau_index[0]:tau_index[1]], label="Estimated $\\beta_{0}$", linestyle="--", linewidth=2)
             
             plt.grid(linestyle='dashed')
             plt.xlabel('$t/n$',fontsize="xx-large")
@@ -1787,8 +1809,8 @@ class LocalLinear:
             plt.figure(figsize=(10, 6 * self.n_est))
             for i in range(self.n_est):
                 plt.subplot(self.n_est, 1, i + 1)
-                plt.plot(x_vals, self.betahat[i],
-                            label=f'Estimated $\\beta_{i}$', color='black')
+                plt.plot(x_vals[tau_index[0]:tau_index[1]], self.betahat[i][tau_index[0]:tau_index[1]],
+                            label=f'Estimated $\\beta_{i}$', linestyle="--", color='black')
 
                 
                 plt.grid(linestyle='dashed')
@@ -1797,17 +1819,42 @@ class LocalLinear:
                 plt.legend(fontsize="x-large")
             plt.show()
     
-    def plot_predicted(self):
+    def plot_predicted(self, tau : list=None):
 
         """
         Plot the actual values of Y against the predicted values of Y over a normalized x-axis from 0 to 1.
+        
+        Parameters
+        ----------
+        tau : list, optional
+            The list looks the  following: tau = [start,end].
+            The function will plot all data and estimates between start and end.
+        
+        Raises
+        ------
+        ValueError
+            No valid tau is provided.
+        
         """
+        tau_index = None
+        if tau is None:
+            tau_index=np.array([0,self.n])
+        elif isinstance(tau, list):
+            if min(tau) <= 0:
+                tau_index = np.array([int(0), int(max(tau) * self.n)])
+            else:
+                tau_index = np.array([int(min(tau)*self.n-1),int(max(tau)*self.n)])
+        else:
+            raise ValueError('The optional parameter tau is required to be a list.')
+            
         plt.figure(figsize=(12, 6))
 
-        x_vals = np.arange(1/self.n,(self.n+1)/self.n,1/self.n)[self.tau_index[0]: self.tau_index[1]]
+        x_vals = np.arange(1/self.n,(self.n+1)/self.n,1/self.n)
+        
+        
 
-        plt.plot(x_vals, self.vY[self.tau_index[0]: self.tau_index[1]], label="True data", linewidth=2,color='black')
-        plt.plot(x_vals, self.predicted_y, label="Fit", linewidth=2)        
+        plt.plot(x_vals[tau_index[0]:tau_index[1]], self.vY[tau_index[0]:tau_index[1]], label="True data", linewidth=1,color='black')
+        plt.plot(x_vals[tau_index[0]:tau_index[1]], self.predicted_y[tau_index[0]:tau_index[1]], label="Fit", linestyle="--",  linewidth=2)        
         plt.grid(linestyle='dashed')
         plt.xlabel('$t/n$',fontsize="xx-large")
         plt.tick_params(axis='both', labelsize=16)
@@ -1815,10 +1862,21 @@ class LocalLinear:
         
         plt.show()
 
-    def plot_residuals(self):
+    def plot_residuals(self, tau : list=None):
         '''
         Plot the residuals over a normalized x-axis from 0 to 1.
-
+        
+        Parameters
+        ----------
+        tau : list, optional
+            The list looks the  following: tau = [start,end].
+            The function will plot all data and estimates between start and end.
+        
+        Raises
+        ------
+        ValueError
+            No valid tau is provided.
+        
         Returns
         -------
         self.residuals : np.ndarray
@@ -1827,9 +1885,19 @@ class LocalLinear:
         '''
         plt.figure(figsize=(12, 6))
 
-        x_vals = np.arange(1/self.n,(self.n+1)/self.n,1/self.n)[self.tau_index[0]: self.tau_index[1]]
+        x_vals = np.arange(1/self.n,(self.n+1)/self.n,1/self.n)
+        tau_index = None
+        if tau is None:
+            tau_index=np.array([0,self.n])
+        elif isinstance(tau, list):
+            if min(tau) <= 0:
+                tau_index = np.array([int(0), int(max(tau) * self.n)])
+            else:
+                tau_index = np.array([int(min(tau)*self.n-1),int(max(tau)*self.n)])
+        else:
+            raise ValueError('The optional parameter tau is required to be a list.')
 
-        plt.plot(x_vals, self.residuals, label="Residuals")       
+        plt.plot(x_vals[tau_index[0]:tau_index[1]], self.residuals[tau_index[0]:tau_index[1]],  linestyle="--", label="Residuals")       
         plt.grid(linestyle='dashed')
         plt.xlabel('$t/n$',fontsize="xx-large")
         plt.tick_params(axis='both', labelsize=16)
@@ -1887,7 +1955,7 @@ class LocalLinear:
         
         # Construct confidence bands
         if Gsubs is None:
-            confidence_bands_list = self.construct_confidence_bands(
+            confidence_bands_list = self._construct_confidence_bands(
                 bootstrap_type, alpha=alpha, gamma=gamma, ic=ic, Chtilde=Chtilde, B=B)
             confidence_bands = confidence_bands_list[:-1][0]
             betahat = confidence_bands_list[-1]
@@ -1897,7 +1965,7 @@ class LocalLinear:
             P_UB = confidence_bands[3]
             G_full = np.linspace(0, 1, len(self.vY))
         else:
-            confidence_bands_dict = self.construct_confidence_bands(
+            confidence_bands_dict = self._construct_confidence_bands(
                 bootstrap_type, alpha=alpha, gamma=gamma, ic=ic, Gsubs=Gsubs, Chtilde=Chtilde, B=B)
             confidence_bands_list = confidence_bands_dict[:-1]
             betahat = confidence_bands_dict[-1]
